# tests for sqlitekv.py

import tempfile

import pytest

from geneva.utils.sqlitekv import SQLiteKV


def test_sqlitekv() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        kv = SQLiteKV(db_path=f"{tmpdir}/kv.sqlite")
        kv["key"] = "value"
        assert kv["key"] == "value"
        assert len(kv) == 1
        assert list(kv) == ["key"]
        del kv["key"]
        assert len(kv) == 0
        with pytest.raises(KeyError):
            kv["key"]


def test_sqlitekv_disk_roundtrip() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        kv = SQLiteKV(db_path=f"{tmpdir}/kv.sqlite")
        kv["key"] = "value"

        # reopens the database
        kv = SQLiteKV(db_path=f"{tmpdir}/kv.sqlite")
        assert kv["key"] == "value"
        assert len(kv) == 1
        assert list(kv) == ["key"]
        del kv["key"]
        assert len(kv) == 0
        with pytest.raises(KeyError):
            kv["key"]


def test_sqlitekv_duplicated_handle() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        kv1 = SQLiteKV(db_path=f"{tmpdir}/kv.sqlite")
        kv2 = SQLiteKV(db_path=f"{tmpdir}/kv.sqlite")

        kv1["key"] = "value"
        assert kv2["key"] == "value"
        del kv1["key"]
        with pytest.raises(KeyError):
            kv2["key"]



import json
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any

import pyarrow as pa

# Self is not available in python 3.10
from typing_extensions import Self, override  # noqa: UP035

from geneva.db import connect
from geneva.table import Table

if TYPE_CHECKING:
    from lance import LanceDataset


class Query(ABC):
    """Present a Query"""

    @abstractmethod
    def to_batches(self, batch_size: int | None = None) -> Iterator[pa.RecordBatch]: ...

    @property
    @abstractmethod
    def schema(self) -> pa.Schema: ...

    @abstractmethod
    def to_dict(self) -> dict: ...

    @staticmethod
    def from_dict(data: dict[str, Any] | str) -> "Query":
        if isinstance(data, str):
            data: dict = json.loads(data)

        match data["type"]:
            case "scan":
                db = connect(data["db"])
                table = db.open_table(data["table"])
                return Scan(table, columns=data["columns"])
            case "projection":
                return Projection(
                    Query.from_dict(data["source"]),
                    columns=[Column(name) for name in data["columns"]],
                )
            case "filter":
                return Filter(Query.from_dict(data["source"]), data["filter"])
            case _:
                raise ValueError(f"Unknown query type: {data['type']}")

    def filter(self, filter: str) -> "Query":  # noqa: A002
        return Filter(self, filter)

    def cache(self) -> Self:
        return self

    def shuffle(self) -> Self:
        return self


class Column:
    """Present a Column in the Table."""

    def __init__(self, name: str) -> None:
        """Define a column."""
        self.name = name

    def alias(self, alias: str) -> "Column":
        return AliasColumn(self, alias)

    def blob(self) -> "Column":
        return BlobColumn(self)

    def apply(self, batch: pa.RecordBatch) -> tuple[str, pa.Array]:
        return (self.name, batch[self.name])


class BlobColumn(Column):
    def __init__(self, col: Column) -> None:
        self.inner = col


class AliasColumn(Column):
    def __init__(self, col: Column, alias: str) -> None:
        self.col = col
        self._alias = alias

    def apply(self, batch: pa.RecordBatch) -> tuple[str, pa.Array]:
        _, arr = self.col.apply(batch)
        return (self._alias, arr)


class Projection(Query):
    def __init__(self, source: Query, columns: list[Column]) -> None:
        self.source = source
        self.columns = columns
        col_names = [col.name for col in columns]
        if col_names:
            self.source = source.with_columns(col_names)

    def with_columns(self, columns: list[str]) -> Self:
        return Projection(self.source, [Column(name) for name in columns])

    def with_row_id(self) -> Self:
        self.source = self.source.with_row_id()
        return self

    def take_rows(self, rows: list[int]) -> pa.Table:
        return self.source.take_rows(rows)

    @override
    def to_dict(self) -> dict:
        return {
            "type": "projection",
            "source": self.source.to_dict(),
            "columns": [col.name for col in self.columns],
        }

    @property
    def schema(self) -> pa.Schema:
        return self.source.schema

    @override
    def to_batches(
        self,
        batch_size: int | None = None,
    ) -> Iterator[dict[str, pa.Array] | pa.RecordBatch]:
        yield from self.source.to_batches(batch_size)


class Filter(Query):
    def __init__(self, source: Query, filter: str) -> None:  # noqa: A002
        self.source = source
        self._filter = filter

    @property
    def schema(self) -> pa.Schema:
        return self.source.schema

    def to_dict(self) -> dict:
        return {
            "type": "filter",
            "source": self.source.to_dict(),
            "filter": self._filter,
        }

    def to_batches(self, batch_size: int | None = None) -> Iterator[pa.RecordBatch]:
        self.source.source._filter = self._filter
        yield from self.source.to_batches(batch_size)


class Scan(Query):
    def __init__(
        self,
        table: Table,
        columns: list[str] | None = None,
        with_row_id: bool = False,
        filter: str | None = None,  # noqa: A002
    ) -> None:
        self.table = table
        self._columns = columns
        self._with_row_id = with_row_id
        self._filter = filter
        # Which columns are blobs?
        self._blob_columns: list[str] = []
        self._offset: int | None = None
        self._limit: int | None = None
        self._fragment_ids: list[int] | None = None

    @classmethod
    def from_uri(cls, db: str, name: str) -> Self:
        table = connect(db).open_table(name)
        return cls(
            table=table,
        )

    def with_filter(self, filter: str | None) -> Self:  # noqa: A002
        self._filter = filter
        return self

    def with_columns(self, columns: list[str]) -> Self:
        self._columns = columns
        return self

    def with_row_id(self) -> Self:
        self._with_row_id = True
        return self

    def with_offset(self, offset: int) -> Self:
        self._offset = offset
        return self

    def with_limit(self, limit: int) -> Self:
        self._limit = limit
        return self

    def with_fragments(self, fragments: list[int] | int) -> Self:
        self._fragment_ids = [fragments] if isinstance(fragments, int) else fragments
        return self

    def __repr__(self) -> str:
        return (
            f"Scan({self.table}, columns={self._columns}, "
            f"with_row_id={self._with_row_id}, filter={self._filter})"
        )

    @override
    def to_dict(self) -> dict:
        return {
            "type": "scan",
            "db": str(self.table._conn.uri),
            "table": self.table._name,
            "columns": self._columns,
            "with_row_id": self._with_row_id,
            "filter": self._filter,
            "fragments": self._fragment_ids,
            "offset": self._offset,
            "limit": self._limit,
        }

    def _scan_schema(self) -> pa.Schema:
        original_schema: pa.Schema = self.table.schema
        if self._columns:
            fields = [original_schema.field(col) for col in self._columns]
        else:
            fields = [original_schema.field(name) for name in original_schema.names]
        return pa.schema(fields)

    @property
    def schema(self) -> pa.Schema:
        scan_schema = self._scan_schema()
        if self._with_row_id:
            fields = [scan_schema.field(name) for name in scan_schema.names] + [
                pa.field("_rowid", pa.int64()),
            ]
            return pa.schema(fields)
        return scan_schema

    def take_rows(self, rows: list[int]) -> pa.Table:
        return self.table._ltbl.to_lance()._take_rows(rows, self._columns)

    def _has_blob_columns(self) -> bool:
        schema = self.schema
        for field in schema:
            if field.metadata and field.metadata.get(b"lance-encoding:blob") == b"true":
                return True
        return False

    @override
    def to_batches(self, batch_size: int | None = None) -> Iterator[pa.RecordBatch]:
        if self._has_blob_columns():
            self.with_row_id()

        schema = self._scan_schema()
        blob_columns = []
        normal_columns = []
        for field in schema:
            if field.metadata and field.metadata.get(b"lance-encoding:blob") == b"true":
                blob_columns.append(field.name)
            else:
                normal_columns.append(field.name)
        if len(blob_columns) > 0:
            self.with_row_id()

        if isinstance(self.table._ltbl, Query):
            for batch in self.table._ltbl.to_batches():
                yield batch
            return

        table_handle: LanceDataset = self.table._ltbl.to_lance()
        fragments = None
        if self._fragment_ids:
            fragments = [table_handle.get_fragment(fid) for fid in self._fragment_ids]

        for batch in table_handle.scanner(
            columns=normal_columns,
            with_row_id=self._with_row_id,
            filter=self._filter,
            batch_size=batch_size,
            offset=self._offset,
            limit=self._limit,
            fragments=fragments,
        ).to_batches():
            if blob_columns:
                ret = batch.to_pydict()
                for blob_column in blob_columns:
                    files = table_handle.take_blobs(
                        batch["_rowid"].to_pylist(),
                        blob_columns[0],
                    )
                    ret[blob_column] = files
                yield ret
            else:
                yield batch



# An dict backed by sqlite3

import functools
import logging
import sqlite3
from collections.abc import Callable, Iterator, MutableMapping
from pathlib import Path
from typing import TypeVar

import attrs

_LOG = logging.getLogger(__name__)


T = TypeVar("T")


@attrs.define
class SQLiteKV(MutableMapping[str, str]):
    """An dict backed by sqlite3"""

    # Path to the sqlite3 database
    db_path: Path = attrs.field(converter=lambda x: Path(x).expanduser().resolve())

    _conn: sqlite3.Connection = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        if not self.db_path.exists():
            _LOG.info("Creating sqlite3 database at %s", self.db_path)
            self.db_path.parent.mkdir(parents=True, exist_ok=True)

        self._conn = sqlite3.connect(
            self.db_path,
        )
        self._conn.execute(
            """
            CREATE TABLE IF NOT EXISTS kv (
                key TEXT PRIMARY KEY,
                value TEXT
            )
            """,
        )
        self._conn.commit()

    @staticmethod
    def _tx_safe(func: Callable[..., T]) -> Callable[..., T]:
        @functools.wraps(func)
        def wrapper(self: "SQLiteKV", *args, **kwargs) -> T:
            self._conn.execute("BEGIN")
            try:
                result = func(self, *args, **kwargs)
                self._conn.commit()
                return result
            except Exception:
                self._conn.rollback()
                raise

        return wrapper

    def close(self) -> None:
        self._conn.close()

    def __getitem__(self, key: str) -> str:
        cursor = self._conn.cursor()
        cursor.execute("SELECT value FROM kv WHERE key=?", (key,))
        result = cursor.fetchone()
        if result is None:
            raise KeyError(key)
        return result[0]

    @_tx_safe
    def __setitem__(self, key: str, value: str) -> None:
        self._conn.execute(
            "INSERT OR REPLACE INTO kv (key, value) VALUES (?, ?)",
            (key, value),
        )

    @_tx_safe
    def __delitem__(self, key: str) -> None:
        self._conn.execute("DELETE FROM kv WHERE key=?", (key,))

    def __iter__(self) -> Iterator[str]:
        cursor = self._conn.cursor()
        cursor.execute("SELECT key FROM kv")
        for row in cursor:
            yield row[0]

    def __len__(self) -> int:
        cursor = self._conn.cursor()
        cursor.execute("SELECT COUNT(*) FROM kv")
        return cursor.fetchone()[0]

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, str):
            return False
        cursor = self._conn.cursor()
        cursor.execute("SELECT 1 FROM kv WHERE key=?", (key,))
        return cursor.fetchone() is not None



# checkpointing utils for lancedb

import abc
import logging
import os
from collections.abc import Iterator

import pyarrow as pa
from lance.file import LanceFileReader, LanceFileWriter
from pyarrow import fs

_LOG = logging.getLogger(__name__)


class CheckpointStore(abc.ABC):
    """Abstract class for checkpoint store, which is a key-value store for storing
    pyarrow.RecordBatch objects

    This is a lighter weight version of collections.abc.MutableMapping
    where we don't expose length or deletion operations
    """

    @abc.abstractmethod
    def __contains__(self, item: str) -> bool:
        pass

    @abc.abstractmethod
    def __getitem__(self, item: str) -> pa.RecordBatch:
        pass

    @abc.abstractmethod
    def __setitem__(self, key: str, value: pa.RecordBatch) -> None:
        pass

    @abc.abstractmethod
    def list_keys(self, prefix: str = "") -> Iterator[str]:
        pass

    @classmethod
    def from_uri(cls, uri: str) -> "CheckpointStore":
        """Construct a CheckpointStore from a URI."""
        if uri == "memory":
            return InMemoryCheckpointStore()
        if uri.startswith(("s3://", "gs://", "az://", "file://")):
            return ArrowFsCheckpointStore(uri)
        raise ValueError(f"Invalid checkpoint store uri: {uri}")


class InMemoryCheckpointStore(CheckpointStore):
    """In memory checkpoint store for testing purposes."""

    def __init__(self) -> None:
        self._store = {}

    def __contains__(self, item: str) -> bool:
        return item in self._store

    def __getitem__(self, item: str) -> pa.RecordBatch:
        return self._store[item]

    def __setitem__(self, key: str, value: pa.RecordBatch) -> None:
        self._store[key] = value

    def list_keys(self, prefix: str = "") -> Iterator[str]:
        for key in self._store:
            if key.startswith(prefix):
                yield key


class LanceCheckpointStore(CheckpointStore):
    def __init__(self, root: str) -> None:
        self.fs, self.path = fs.FileSystem.from_uri(root)  # type: ignore
        self.root = root

    def __contains__(self, key: str) -> bool:
        _LOG.debug("contains: %s", key)
        path = os.path.join(self.path, f"{key}.lance")
        info = self.fs.get_file_info(path)
        return info.type != fs.FileType.NotFound

    def __getitem__(self, key: str) -> pa.RecordBatch:
        _LOG.debug("get: %s", key)
        path = os.path.join(self.root, f"{key}.lance")
        reader = LanceFileReader(path)
        return reader.read_all().to_table().combine_chunks().to_batches()[0]

    def __setitem__(self, key: str, value: pa.RecordBatch) -> None:
        _LOG.debug("set: %s", key)
        path = os.path.join(self.root, f"{key}.lance")
        with LanceFileWriter(path, value.schema) as writer:
            writer.write_batch(value)

    def list_keys(self, prefix: str = "") -> Iterator[str]:
        _LOG.debug("list_keys: %s", prefix)
        selector = fs.FileSelector(os.path.join(self.path, prefix))
        for file in self.fs.get_file_info(selector):
            yield file.path.removeprefix(self.path).lstrip("/").rstrip(".lance")


class ArrowFsCheckpointStore(CheckpointStore):
    def __init__(self, root: str) -> None:
        self.fs, self.path = fs.FileSystem.from_uri(root)  # type: ignore
        self.fs.create_dir(self.path)

    def __contains__(self, item: str) -> bool:
        _LOG.debug("contains: %s", item)
        info = self.fs.get_file_info(os.path.join(self.path, item))
        return info.type != fs.FileType.NotFound

    def __getitem__(self, item: str) -> pa.RecordBatch:
        _LOG.debug("get: %s", item)
        with (
            self.fs.open_input_file(os.path.join(self.path, item)) as f,
            pa.RecordBatchFileReader(f) as r,
        ):
            return r.read_all().combine_chunks()

    def __setitem__(self, key: str, value: pa.RecordBatch) -> None:
        _LOG.debug("set: %s", key)
        with (
            self.fs.open_output_stream(os.path.join(self.path, key)) as f,
            pa.RecordBatchFileWriter(
                f,
                value.schema,
            ) as writer,
        ):
            writer.write(value)

    def list_keys(self, prefix: str = "") -> Iterator[str]:
        _LOG.debug("list_keys: %s", prefix)
        selector = fs.FileSelector(os.path.join(self.path, prefix))
        for file in self.fs.get_file_info(selector):
            yield file.path.removeprefix(self.path).lstrip("/")



import hashlib
import logging
from collections.abc import Iterator

import attrs
import lance
import pyarrow as pa

from geneva.checkpoint import (
    CheckpointStore,
)
from geneva.debug.logger import ErrorLogger, NoOpErrorLogger
from geneva.query import Scan
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


@attrs.define
class ReadTask:
    uri: str
    columns: list[str]
    frag_id: int
    offset: int
    limit: int

    filter: str | None = None

    batch_size: int = 1024

    def to_batches(self) -> Iterator[pa.RecordBatch]:
        uri_parts = self.uri.split("/")
        name = ".".join(uri_parts[-1].split(".")[:-1])
        db = "/".join(uri_parts[:-1])
        scan = (
            Scan.from_uri(db, name)
            .with_columns(self.columns)
            .with_fragments([self.frag_id])
            .with_filter(self.filter)
            .with_offset(self.offset)
            .with_limit(self.limit)
        )
        yield from scan.to_batches(self.batch_size)

    def checkpoint_key(self) -> str:
        hasher = hashlib.md5()
        hasher.update(
            f"{self.uri}:{self.frag_id}:{self.offset}:{self.limit}:{self.filter}".encode(),
        )
        return hasher.hexdigest()


@attrs.define
class LanceRecordBatchUDFApplier:
    udfs: dict[str, UDF] = attrs.field()
    checkpoint_store: CheckpointStore = attrs.field()
    error_logger: ErrorLogger = attrs.field(default=NoOpErrorLogger())

    @property
    def output_schema(self) -> pa.Schema:
        return pa.schema(
            [pa.field(name, fn.data_type) for name, fn in self.udfs.items()],
        )

    def _run(self, task: ReadTask) -> dict[str, str]:
        data_key = task.checkpoint_key()
        _LOG.debug("Running task %s", task)
        # track the batch sequence number so we can checkpoint any errors
        # when reproducing locally we can seek to the erroring batch quickly

        # prepare the schema
        fields = []
        for name, fn in self.udfs.items():
            fields.append(pa.field(name, fn.data_type, metadata=fn.field_metadata))

        res = {}
        batch = None
        for name, fn in self.udfs.items():
            checkpoint_key = f"{data_key}:{fn.checkpoint_key}"
            if checkpoint_key in self.checkpoint_store:
                _LOG.info("Using cached result for %s", checkpoint_key)
                res[name] = checkpoint_key
                continue
            arrs = []
            # TODO: add caching for the input data
            for seq, batch in enumerate(task.to_batches()):
                try:
                    arrs.append(fn(batch))
                except Exception as e:
                    self.error_logger.log_error(e, task, fn, seq)
                    raise e

            arr = pa.concat_arrays(arrs)
            self.checkpoint_store[checkpoint_key] = pa.RecordBatch.from_pydict(
                {"data": arr},
                schema=pa.schema([pa.field("data", fn.data_type)]),
            )
            res[name] = checkpoint_key

        return res

    def run(self, task: ReadTask) -> dict[str, str]:
        try:
            return self._run(task)
        except Exception as e:
            logging.exception("Error running task %s: %s", task, e)
            raise RuntimeError(f"Error running task {task}") from e

    def status(self, task: ReadTask) -> dict[str, str]:
        data_key = task.checkpoint_key()
        return {
            name: f"{data_key}:{fn.checkpoint_key}" in self.checkpoint_store
            for name, fn in self.udfs.items()
        }  # type: ignore


def plan_read(
    uri: str,
    columns: list[str] | None = None,
    *,
    read_version: int | None = None,
    batch_size: int = 512,
    filter: str | None = None,  # noqa: A002
) -> Iterator[ReadTask]:
    """Make Plan for Reading Data from a Dataset"""
    if columns is None:
        columns = []
    dataset = lance.dataset(uri)
    if read_version is not None:
        dataset = dataset.checkout_version(read_version)

    for frag in dataset.get_fragments():
        frag_rows = frag.count_rows(filter=filter)
        for offset in range(0, frag_rows, batch_size):
            limit = min(batch_size, frag_rows - offset)
            yield ReadTask(
                uri=uri,
                columns=columns,
                frag_id=frag.fragment_id,
                batch_size=batch_size,
                offset=offset,
                limit=limit,
                filter=filter,
            )



import hashlib
import logging
import os
import subprocess
from pathlib import Path
from typing import Literal

import attrs
import docker
import docker.models
import docker.models.images
import docker.utils
import docker.utils.json_stream
from ._state import PackagingState
from jinja2 import Environment, FileSystemLoader

_CUDA_BASE = "nvidia/cuda:12.4.1-cudnn-runtime-ubuntu22.04"
_CPU_BASE = "ubuntu:24.04"


_LOG = logging.getLogger(__name__)


def _get_base_dockerfile(
    backend: Literal["dataflow"],
    python_tool_chain: Literal["uv"],
    additional_pip_dependencies: list[str] | None = None,
    cuda: bool = False,
) -> str:
    if additional_pip_dependencies is None:
        additional_pip_dependencies = []
    module_location = Path(__file__).parent
    env = Environment(loader=FileSystemLoader(module_location), autoescape=True)
    # Load a template
    template = env.get_template(f"_base_{python_tool_chain}_{backend}.Dockerfile.j2")

    post_install_command = ""

    if additional_pip_dependencies:
        packages = " ".join(additional_pip_dependencies)
        post_install_command = f"RUN pip install {packages}"

    context = {
        "base_image": _CUDA_BASE if cuda else _CPU_BASE,
        "post_install_command": post_install_command,
    }
    return template.render(context)


def _get_incremental_dockerfile(
    backend: Literal["dataflow"],
    python_tool_chain: Literal["uv"],
    incremental_base: str,
) -> str:
    module_location = Path(__file__).parent
    env = Environment(loader=FileSystemLoader(module_location), autoescape=True)
    template = env.get_template(
        f"_incremental_{python_tool_chain}_{backend}.Dockerfile.j2"
    )

    context = {"incremental_base_image": incremental_base}

    return template.render(context)


@attrs.define
class DockerWorkspacePackager:
    """Local workspace docker packager"""

    # Docker client
    client: docker.DockerClient = attrs.field(
        factory=docker.from_env,
        kw_only=True,
    )

    additional_pip_dependencies: list[str] = attrs.field(factory=list, kw_only=True)

    # Path to the packaging cache
    state_store_path: Path = attrs.field(
        default=Path(os.environ.get("XDG_CONFIG_DIR", "~/.cache"))
        / "_geneva/ws_packager",
        converter=Path,
        kw_only=True,
    )

    _state: PackagingState = attrs.field(init=False)

    # support only uv for now
    python_tool_chain: Literal["uv"] = attrs.field(default="uv", kw_only=True)

    def __attrs_post_init__(self) -> None:
        self._state = PackagingState(self.state_store_path)

    def build(
        self,
        image_name: str,
        cuda: bool = False,
        platform: Literal["linux/amd64", "linux/arm64"] = "linux/amd64",
    ) -> docker.models.images.Image:
        """Build a docker image"""

        # this context determines which base image (if there is one)
        # to use for incremental builds
        #
        # sort the item tuples so that the hash is deterministic
        # from potential order changes
        built_context = sorted(
            {
                "platform": platform,
                "cuda": cuda,
                "backend": "dataflow",
                "additional_pip_dependencies": self.additional_pip_dependencies,
                "python_tool_chain": self.python_tool_chain,
            }.items()
        )

        context_hash = hashlib.md5(str(built_context).encode()).hexdigest()

        if (
            incremental_image := self._state.current_incremental_image(context_hash)
        ) is not None:
            dockerfile = _get_incremental_dockerfile(
                backend="dataflow",
                python_tool_chain=self.python_tool_chain,
                incremental_base=incremental_image,
            )
            is_incremental = True
        else:
            dockerfile = _get_base_dockerfile(
                backend="dataflow",
                python_tool_chain=self.python_tool_chain,
                additional_pip_dependencies=self.additional_pip_dependencies,
                cuda=cuda,
            )
            is_incremental = False

        subprocess.run(
            [
                "docker",
                "build",
                "-f",
                "-",
                "-t",
                image_name,
                "--platform",
                platform,
                ".",
            ],
            input=dockerfile.encode(),
            env={"DOCKER_BUILDKIT": "1", **os.environ.copy()},
            check=True,
        )

        # avoid setting incremental image on incremental builds
        # this is because the last layer of the incremental image
        # could be very large and overlaps with the incoming changes
        # TODO: we should trigger a full build of the incremental
        # base image based on some heuristic, e.g. change size, deps, etc.
        if not is_incremental:
            self._state.set_incremental_image(context_hash, image_name)

        return self.client.images.get(image_name)

    def push(self, image_name: str) -> None:
        """Push a docker image"""
        self.client.images.push(image_name)


def register_packager_parser(subparsers) -> None:
    packager_parser = subparsers.add_parser(
        "package", description="Build local workspace and push docker image"
    )
    packager_parser.add_argument(
        "--docker-image",
        dest="image_name",
        required=True,
        help="The name of the image to build.",
    )
    packager_parser.add_argument(
        "--cuda",
        dest="cuda",
        action="store_true",
        default=False,
        help="Use CUDA base image.",
    )
    packager_parser.add_argument(
        "--platform",
        dest="platform",
        choices=["linux/amd64", "linux/arm64"],
        help="The platform to build the image for.",
    )
    packager_parser.set_defaults(func=run)


def run(args) -> None:
    packager = DockerWorkspacePackager()
    image = packager.build(
        image_name=args.image_name,
        cuda=args.cuda,
        platform=args.platform,
    )

    _LOG.info("Built image %s, pushing to %s", image.id, args.image_name)
    packager.client.images.push(args.image_name)
    _LOG.info("Pushed image %s", args.image_name)



from pathlib import Path

import lance
import pyarrow as pa

from geneva import LanceCheckpointStore, udf
from geneva.runners.ray.pipeline import run_ray_add_column


def test_run_ray_add_column(tmp_path: Path) -> None:
    @udf(data_type=pa.int32(), batch_size=8)
    def add_one(a) -> int:
        return a + 1

    data = {"a": pa.array(range(100))}
    tbl = pa.Table.from_pydict(data)
    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    tbl_path = tmp_path / "foo.lance"
    ds = lance.write_dataset(tbl, tbl_path)

    def return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
        return pa.RecordBatch.from_pydict(
            {"b": pa.array([None] * batch.num_rows, pa.int32())}
        )

    new_fragment, new_schema = ds.get_fragment(0).merge_columns(
        return_none, columns=["a"]
    )

    merge = lance.LanceOperation.Merge([new_fragment], new_schema)
    lance.LanceDataset.commit(tbl_path, merge, read_version=ds.version)

    run_ray_add_column(str(tbl_path), ["a"], {"b": add_one}, ckp_store)


def test_ray_with_batch_udf(tmp_path: Path) -> None:
    pass



from lancedb.common import Credential
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.theme import BUILTIN_THEMES
from textual.widgets import Button, Footer, Header, Input, ProgressBar, Select, Static

import geneva
from geneva.db import Connection
from geneva.tui.error import show_error
from geneva.tui.tables import TablesList
from geneva.utils.sqlitekv import SQLiteKV


class ThemeSelector(Select[str]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            [(theme, theme) for theme in BUILTIN_THEMES],
            *args,
            prompt="Theme",
            allow_blank=False,
            **kwargs,
        )

    def _watch_value(self, value) -> None:
        super()._watch_value(value)
        self.app.theme = value


class DbConnectionConfigBlock(Vertical):
    def __init__(self, *args, state: SQLiteKV, **kwargs) -> None:
        self._state = state

        self._uri = Input(
            placeholder="URI", id="uri", value=self._state.get("uri", None)
        )
        self._region = Input(
            placeholder="Region", id="region", value=self._state.get("region", None)
        )
        self._api_key = Input(placeholder="API Key", id="api_key", password=True)
        self._host_override = Input(
            placeholder="Host Override",
            id="host_override",
            value=self._state.get("host_override", None),
        )

        super().__init__(
            self._uri,
            self._region,
            self._api_key,
            self._host_override,
            *args,
            **kwargs,
        )

    def get_connection(self) -> Connection:
        if not self._uri.value:
            raise ValueError("URI is required.")
        if not self._region.value:
            raise ValueError("Region is required.")
        if not self._api_key.value:
            raise ValueError("API Key is required.")
        if not self._host_override.value:
            raise ValueError("Host Override is required.")

        self._state["uri"] = self._uri.value
        self._state["region"] = self._region.value
        self._state["host_override"] = self._host_override.value

        conn = geneva.connect(
            uri=self._uri.value,
            region=self._region.value,
            api_key=Credential(self._api_key.value),
            host_override=self._host_override.value,
        )
        return conn

    DEFAULT_CSS = """
    DbConnectionConfigBlock {
        width: 80%;
    }
    """


class GenevaLake(App):
    def __init__(
        self,
        *args,
        db_connection: Connection | None = None,
        error: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.db_connection = db_connection
        self.error = error
        self._state = SQLiteKV(".geneva_tui.db")

    def compose(self) -> ComposeResult:
        yield Header(name="geneva")
        theme_selector = ThemeSelector(
            value=next(iter(BUILTIN_THEMES)),
        )
        theme_selector_group = Vertical(
            Static("Theme", id="theme_selector_label"),
            theme_selector,
            id="theme_selector_group",
        )

        self.connection_config = DbConnectionConfigBlock(state=self._state)
        reconnect_button = Button("Connect/Refresh", id="connect_button")
        connection_config_group = Vertical(
            Static("DB Connection", id="db_connection_label"),
            self.connection_config,
            reconnect_button,
            id="connection_config_group",
        )

        separator = ProgressBar(
            show_bar=True,
            show_eta=False,
            show_percentage=False,
            id="setting_title_separator",
        )
        separator.total = 100
        separator.progress = 100
        setting_group = Vertical(
            Static("Settings", id="settings_label"),
            separator,
            theme_selector_group,
            connection_config_group,
            id="setting_group",
        )

        self.table_list = TablesList()
        self.table_list.set_connection(self.db_connection)

        db_tree_group = Vertical(
            Static("DB Tree", id="db_tree_label"),
            self.table_list,
            id="db_tree_group",
        )

        yield Horizontal(
            db_tree_group,
            setting_group,
        )
        yield Footer()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "connect_button":
            try:
                conn = self.connection_config.get_connection()
                self.table_list.set_connection(conn)
            except Exception as e:
                show_error(self.app, ex=e)
                self.table_list.set_connection(None)

    CSS = """
    #setting_group {
        width: 50vw
    }
    #theme_selector_label {
        text-align: center;
    }
    #settings_label {
        text-align: center;
    }
    #db_tree_label {
        text-align: center;
    }
    #setting_title_separator {
        align: center middle;
        width: 100%;
    }
    #db_connection_label {
        text-align: center;
    }
    #connection_config_group {
        align: center top;
    }
    #connect_button {
        align: center middle;
    }
    """


if __name__ == "__main__":
    GenevaLake().run()



import base64
import logging
import os
import uuid
from collections.abc import Iterable, Iterator
from datetime import timedelta
from functools import cached_property
from typing import Any, Literal

import cloudpickle
import lance
import lancedb
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
from lance.schema import schema_to_json
from lancedb.common import VECTOR_COLUMN_NAME
from lancedb.index import IndexConfig
from lancedb.query import LanceQueryBuilder
from lancedb.query import Query as LanceQuery
from lancedb.table import IndexStatistics
from lancedb.table import Table as LanceTable

# Python 3.10 compatibility
from typing_extensions import Never, override  # noqa: UP035

from geneva.db import Connection
from geneva.docker.packager import DockerWorkspacePackager
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


class Table(LanceTable):
    """Table in Geneva.

    A Table is a Lance dataset
    """

    def __init__(
        self,
        conn: Connection,
        name: str,
        *,
        version: int | None = None,
        storage_options: dict[str, str] | None = None,
        index_cache_size: int | None = None,
        **kwargs,
    ) -> None:
        self._conn_uri = conn.uri
        self._name = name

        self._conn = conn

        base_uri = conn._uri.removesuffix("/")
        self._uri = f"{base_uri}/{name}.lance"
        self._version: int | None = version
        self._index_cache_size = index_cache_size
        self._storage_options = storage_options

        # Load table
        self._ltbl  # noqa

    def __repr__(self) -> str:
        return f"<Table {self._name}>"

    # TODO: This annotation sucks
    def __reduce__(self):  # noqa: ANN204
        return (self.__class__, (self._conn, self._name))

    def get_fragments(self) -> list[lance.LanceFragment]:
        return self._ltbl.to_lance().get_fragments()

    @cached_property
    def _ltbl(self) -> lancedb.table.Table:
        # remote db, open table directly
        if self._conn_uri.startswith("db://"):
            tbl = self._conn._connect.open_table(self._name)
            if self._version:
                tbl.checkout(self._version)
            return tbl

        # object store db, open table from file
        view_file_path = os.path.join(self._uri, "view.json")
        fs, path = pa.fs.FileSystem.from_uri(view_file_path)
        file_info = fs.get_file_info(path)
        if file_info.type == pa.fs.FileType.NotFound:
            tbl = self._conn._connect.open_table(self._name)
            if self._version:
                tbl.checkout(self._version)
            return tbl

        fs, parent_dir = pa.fs.FileSystem.from_uri(self._uri)
        fs.create_dir(parent_dir, recursive=True)
        with fs.open_input_file(path) as f:
            q_json = f.read().decode("utf-8")
            from .query import Query

            return Query.from_dict(q_json)

    @property
    def name(self) -> str:
        """Get the name of the table."""
        return self._name

    @property
    def version(self) -> int:
        """Get the current version of the table"""
        return self._ltbl.version

    @property
    def schema(self) -> pa.Schema:
        """The Arrow Schema of the Table."""
        return self._ltbl.schema

    @property
    def uri(self) -> str:
        return self._uri

    @property
    def embedding_functions(self) -> Never:
        raise NotImplementedError("Embedding functions are not supported.")

    def add(
        self,
        data,
        mode: str = "append",
        on_bad_vectors: str = "error",
        fill_value: float = 0.0,
    ) -> None:
        self._ltbl.add(
            data,
            mode=mode,
            on_bad_vectors=on_bad_vectors,
            fill_value=fill_value,
        )

    def checkout(self, version: int) -> None:
        self._version = version
        self._ltbl.checkout(version)

    def checkout_latest(self) -> None:
        self._ltbl.checkout_latest()

    def add_columns(
        self,
        mapping: dict[str, UDF],
        *,
        input_columns: list[str] | None = None,
        **kwargs,
    ) -> None:
        if len(mapping) != 1:
            raise ValueError("Only one UDF is supported for now.")

        _LOG.info("Adding column: udf=%s", mapping)
        col_name = next(iter(mapping))
        udf = mapping[col_name]
        image_tag, udf_pickle = self.package_udf(col_name, udf)

        # TODO: Add a local fireup test to make sure:
        # 1. the docker and the pickle and load
        # 2. when a batch is passed, the UDF can be executed correctly

        if not input_columns:
            input_columns = udf.input_columns

        client = self._conn._client
        data_type = schema_to_json(pa.schema({"col": udf.data_type}))["fields"][0][
            "type"
        ]
        req_data = {
            "new_columns": [
                {
                    "name": col_name,
                    "virtual_column": {
                        "image": image_tag,
                        "udf": base64.b64encode(udf_pickle).decode("utf-8"),
                        "data_type": data_type,
                        "input_columns": input_columns,
                        "udf_version": udf.version,
                        "udf_name": udf.name,
                    },
                }
            ]
        }
        result = client.post(
            f"v1/table/{self.name}/add_columns/",
            data=req_data,
            content_type="application/json",
        )

        _LOG.debug("Add columns result: %s", result)

    def alter_columns(self, *alterations: Iterable[dict[str, Any]]) -> None:
        basic_column_alterations = []
        virtual_column_alterations = []

        for alter in alterations:
            if "virtual_column" in alter:
                if "path" not in alter:
                    raise ValueError("path is required to alter virtual virtual_column")
                if not isinstance(alter["virtual_column"], UDF):
                    raise ValueError("virtual_column must be a UDF")

                col_name = alter["path"]
                udf = alter["virtual_column"]
                image_tag, udf_pickle = self.package_udf(col_name, udf)

                virtual_column_alterations.append(
                    {
                        "path": col_name,
                        "virtual_column": {
                            "image": image_tag,
                            "udf": base64.b64encode(udf_pickle).decode("utf-8"),
                            "udf_version": udf.version,
                            "udf_name": udf.name,
                            "input_columns": udf.input_columns,
                        },
                    }
                )
            else:
                basic_column_alterations.append(alter)

        if len(virtual_column_alterations) > 0:
            client = self._conn._client
            req_data = {"alterations": virtual_column_alterations}
            result = client.post(
                f"v1/table/{self.name}/alter_columns/",
                data=req_data,
                content_type="application/json",
            )
            _LOG.debug("Alter virtual columns result: %s", result)

        if len(basic_column_alterations) > 0:
            self._ltbl.alter_columns(*basic_column_alterations)

    def package_udf(self, col_name: str, udf: UDF) -> tuple[str, bytes]:
        udf_id = uuid.uuid4().hex

        # TODO(rmeng): this should come from the client or config
        _image_prefix = "us-docker.pkg.dev/lancedb-devland/dataflow-devland/lance-beam"
        packager = DockerWorkspacePackager()
        image_tag = f"{_image_prefix}:{udf_id}"
        _LOG.info("Building UDF image: %s", image_tag)
        packager.build(image_tag, platform="linux/amd64", cuda=udf.cuda)
        _LOG.info("Pushing UDF image: %s", image_tag)
        packager.push(image_tag)
        _LOG.info("Pushed UDF image: %s", image_tag)

        udf_pickle = cloudpickle.dumps(udf)

        return (image_tag, udf_pickle)

    def create_index(
        self,
        metric: str = "L2",
        num_partitions: int | None = None,
        num_sub_vectors: int | None = None,
        vector_column_name: str = VECTOR_COLUMN_NAME,
        replace: bool = True,
        accelerator=None,
        index_cache_size=None,
        *,
        index_type: Literal[
            "IVF_FLAT",
            "IVF_PQ",
            "IVF_HNSW_SQ",
            "IVF_HNSW_PQ",
        ] = "IVF_PQ",
        num_bits: int = 8,
        max_iterations: int = 50,
        sample_rate: int = 256,
        m: int = 20,
        ef_construction: int = 300,
    ) -> None:
        """Create Vector Index"""
        self._ltbl.create_index(
            metric,
            num_partitions or 256,
            num_sub_vectors or 96,
            vector_column_name,
            replace,
            accelerator,
            index_cache_size,
            index_type=index_type,
            num_bits=num_bits,
            max_iterations=max_iterations,
            sample_rate=sample_rate,
            m=m,
            ef_construction=ef_construction,
        )

    @override
    def create_fts_index(
        self,
        field_names: str | list[str],
        ordering_field_names: str | list[str] | None = None,
        *,
        replace: bool = False,
        writer_heap_size: int | None = None,
        tokenizer_name: str | None = None,
        with_position: bool = True,
        base_tokenizer: Literal["simple", "raw", "whitespace"] = "simple",
        language: str = "English",
        max_token_length: int | None = 40,
        lower_case: bool = True,
        stem: bool = False,
        remove_stop_words: bool = False,
        ascii_folding: bool = False,
        **_kwargs,
    ) -> None:
        self._ltbl.create_fts_index(
            field_names,
            ordering_field_names,  # type: ignore
            replace=replace,
            writer_heap_size=writer_heap_size,
            tokenizer_name=tokenizer_name,
            with_position=with_position,
            base_tokenizer=base_tokenizer,
            language=language,
            max_token_length=max_token_length,
            lower_case=lower_case,
            stem=stem,
            remove_stop_words=remove_stop_words,
            ascii_folding=ascii_folding,
        )

    @override
    def create_scalar_index(
        self,
        column: str,
        *,
        replace: bool = True,
        index_type: Literal["BTREE", "BITMAP", "LABEL_LIST"] = "BTREE",
    ) -> None:
        self._ltbl.create_scalar_index(
            column,
            replace=replace,
            index_type=index_type,
        )

    @override
    def _do_merge(self, other: "Table", on: list[str], how: str) -> Never:
        raise NotImplementedError("Merging tables is not supported.")

    @override
    def _execute_query(
        self,
        query: LanceQuery,
        batch_size: int | None = None,
    ) -> pa.RecordBatchReader:
        return self._ltbl._execute_query(query, batch_size=batch_size)

    def list_versions(self) -> list[dict[str, Any]]:
        return self._ltbl.list_versions()

    @override
    def cleanup_old_versions(
        self,
        older_than: timedelta | None = None,
        *,
        delete_unverified=False,
    ) -> "lance.CleanupStats":
        return self._ltbl.cleanup_old_versions(
            older_than,
            delete_unverified=delete_unverified,
        )

    def to_batches(self, batch_size: int | None = None) -> Iterator[pa.RecordBatch]:
        from .query import Query

        if isinstance(self._ltbl, Query):
            return self._ltbl.to_batches(batch_size)
        return self._ltbl.to_lance().to_batches(batch_size)

    def scanner(
        self,
        columns: list[str] | dict[str, str] | None = None,
        filter: str | pc.Expression | None = None,  # noqa: A002
        limit: int | None = None,
        offset: int | None = None,
        batch_size: int | None = None,
        batch_readahead: int | None = None,
        fragment_readahead: int | None = None,
        scan_in_order: bool | None = None,
        *,
        with_row_id: bool | None = None,
        with_row_address: bool | None = None,
        use_stats: bool | None = None,
        fast_search: bool | None = None,
        io_buffer_size: int | None = None,
        late_materialization: bool | list[str] | None = None,
        use_scalar_index: bool | None = None,
        **kwargs,
    ) -> lance.LanceScanner | Iterator[pa.RecordBatch]:
        from .query import Query

        if isinstance(self._ltbl, Query):
            return self._ltbl.to_batches()

        return self._ltbl.to_lance().scanner(
            columns=columns,
            filter=filter,
            limit=limit,
            offset=offset,
            batch_size=batch_size,
            batch_readahead=batch_readahead,
            fragment_readahead=fragment_readahead,
            scan_in_order=scan_in_order,
            with_row_id=with_row_id,
            with_row_address=with_row_address,
            use_stats=use_stats,
            fast_search=fast_search,
            io_buffer_size=io_buffer_size,
            late_materialization=late_materialization,
            use_scalar_index=use_scalar_index,
        )

    def search(
        self,
        query: list | pa.Array | pa.ChunkedArray | np.ndarray | None = None,
        vector_column_name: str | None = None,
        query_type: Literal["vector", "fts", "hybrid", "auto"] = "auto",
        ordering_field_name: str | None = None,
        fts_columns: str | list[str] | None = None,
    ) -> LanceQueryBuilder:
        return self._ltbl.search(
            query=query,
            vector_column_name=vector_column_name,
            query_type=query_type,
            fts_columns=fts_columns,
        )

    @override
    def drop_columns(self, columns: Iterable[str]) -> None:
        self._ltbl.drop_columns(columns)

    @override
    def to_arrow(self) -> pa.Table:
        return self._ltbl.to_arrow()

    @override
    def count_rows(self, filter: str | None = None) -> int:
        return self._ltbl.count_rows(filter)

    @override
    def update(
        self,
        where: str | None = None,
        values: dict | None = None,
        *,
        values_sql: dict[str, str] | None = None,
    ) -> None:
        self._ltbl.update(where, values, values_sql=values_sql)

    @override
    def delete(self, where: str) -> None:
        self._ltbl.delete(where)

    @override
    def list_indices(self) -> Iterable[IndexConfig]:
        return self._ltbl.list_indices()

    @override
    def index_stats(self, index_name: str) -> IndexStatistics | None:
        return self._ltbl.index_stats(index_name)

    @override
    def optimize(
        self,
        *,
        cleanup_older_than: timedelta | None = None,
        delete_unverified: bool = False,
    ) -> None:
        return self._ltbl.optimize(
            cleanup_older_than=cleanup_older_than,
            delete_unverified=delete_unverified,
        )

    @override
    def compact_files(self) -> None:
        self._ltbl.compact_files()

    # TODO: This annotation sucks
    def take_blobs(self, indices: list[int] | pa.Array, column: str):  # noqa: ANN201
        return self._ltbl.to_lance().take_blobs(indices, column)

    def uses_v2_manifest_paths(self) -> bool:
        return self._ltbl.uses_v2_manifest_paths()

    def migrate_v2_manifest_paths(self) -> None:
        return self._ltbl.migrate_v2_manifest_paths()



import json

import cloudpickle
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.widgets import Button, DataTable, Footer, Header, Static

from geneva.apply import ReadTask
from geneva.debug.entrypoint import run_udf_task_with_debug
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.tui._screen import DismissOnQuitScreen
from geneva.tui.error import show_error


class ErrorLogsScreen(DismissOnQuitScreen):
    def __init__(
        self,
        *args,
        error_logger: CheckpointStoreErrorLogger,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.title = "Error Logs"
        self.error_logger = error_logger

    def compose(self) -> ComposeResult:
        yield Header(name=self.title)
        yield Footer()
        message = ""
        buttons = Vertical(
            Button("Connect to Debugger", id="debug"),
            Button("Run in Profiler", id="profile"),
            id="buttons",
        )

        self.table = get_error_table(self.error_logger)
        self.table.id = "error_table"

        yield Vertical(
            self.table,
            Horizontal(
                buttons,
                VerticalScroll(Static(message)),
            ),
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """An action to select the cursor."""

        row = self.table.cursor_row
        row = self.table.get_row_at(row)
        key = row[0]

        try:
            error_row = self.error_logger.get_error_row(key)
            task_json = error_row["task"][0].as_py()
            udf_data = error_row["udf"][0].as_py()

            task_dict = json.loads(task_json)
            task = ReadTask(**task_dict)
            udf = cloudpickle.loads(udf_data)
        except Exception as e:
            show_error(self.app, ex=e)
            return

        with self.app.suspend():
            run_udf_task_with_debug(task, udf)

    CSS = """
    #buttons {
        width: 10vw;
    }
    #error_table {
        height: 80vh;
        width: 100vw;
    }
    """


def get_error_table(
    error_logger: CheckpointStoreErrorLogger,
) -> DataTable:
    error_keys = list(error_logger.list_errors())

    table = DataTable(cursor_type="row")
    table.add_columns(
        "ID",
        "URI",
        "Input Columns",
        "Frament ID",
        "Offset",
        "Limit",
        "Batch Size",
        "Error",
    )

    for key in error_keys:
        error_row = error_logger.get_error_row(key)
        task_json = error_row["task"][0].as_py()
        task_dict = json.loads(task_json)
        error = error_row["error"][0].as_py()

        table.add_row(
            key,
            task_dict["uri"],
            task_dict["columns"],
            task_dict["frag_id"],
            task_dict["offset"],
            task_dict["limit"],
            task_dict["batch_size"],
            error,
        )

    return table



from more_itertools import peekable
from textual.app import ComposeResult
from textual.widgets import DataTable, Footer, Header, Static
from textual.widgets.data_table import ColumnKey

from geneva.checkpoint import ArrowFsCheckpointStore
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.job.client import JobClient
from geneva.tui._screen import DismissOnQuitScreen
from geneva.tui.error import show_error
from geneva.tui.error_logs import ErrorLogsScreen


class JobsTable(DataTable):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, cursor_type="row", **kwargs)

    def action_select_cursor(self) -> None:
        row = self.cursor_row
        row = self.get_row_at(row)

        job_id_column_idx = self._column_locations.get(ColumnKey(value="id"))
        store_column_idx = self._column_locations.get(
            ColumnKey(value="checkpoint_store")
        )

        if job_id_column_idx is None:
            show_error(self.app, msg="No job id column found, this is not expected")
            return

        if store_column_idx is None:
            show_error(
                self.app, msg="No checkpoint store column found, this is not expected"
            )
            return

        store_name = row[store_column_idx]
        job_id = row[job_id_column_idx]

        checkpoint_store = ArrowFsCheckpointStore(store_name)
        error_logger = CheckpointStoreErrorLogger(
            job_id=job_id, checkpoint_store=checkpoint_store
        )

        error_log_screen = ErrorLogsScreen(error_logger=error_logger)
        self.app.push_screen(error_log_screen)


class JobsScreen(DismissOnQuitScreen):
    def __init__(
        self, *args, client: JobClient, table_name: str | None = None, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.title = "Jobs"
        self.table_name = table_name
        self.client = client

    def compose(self) -> ComposeResult:
        yield Header(name=self.title)
        yield Footer()
        yield Static("Job History", id="job_history_label")

        table = JobsTable()

        try:
            first_job = peekable(self.client.list(table=self.table_name))
            for key in first_job.peek():
                table.add_column(label=key, key=key)

            # TODO(rmeng): make the UI partially load before the data is fetched
            for job in first_job:
                table.add_row(*job.values())

            yield table
        except StopIteration:
            yield Static("No jobs found")



# Test data lake works with PyTorch

from pathlib import Path

import lance
import lance.torch.data
import pyarrow as pa
import torch

from geneva import connect


def test_torch_dataset(tmp_path: Path) -> None:
    db = connect(tmp_path)
    data = pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})

    tbl = db.create_table("test", data)

    ds = lance.torch.data.LanceDataset(tbl, batch_size=2)

    tensor = next(iter(ds))
    assert torch.equal(tensor["a"], torch.tensor([1, 2]))
    assert torch.equal(tensor["b"], torch.tensor([4, 5]))



# a error logger for UDF jobs
import abc
import json
from collections.abc import Iterator
from typing import TYPE_CHECKING

import attrs
import cloudpickle
import pyarrow as pa

from geneva.checkpoint import CheckpointStore

if TYPE_CHECKING:
    from geneva.apply import ReadTask
    from geneva.transformer import UDF


class ErrorLogger(abc.ABC):
    @abc.abstractmethod
    def log_error(
        self,
        error: Exception,
        task: "ReadTask",
        udf: "UDF",
        seq: int,
    ) -> None: ...

    @abc.abstractmethod
    def list_errors(self) -> Iterator[str]: ...

    @abc.abstractmethod
    def get_error_row(self, error_id: str) -> pa.RecordBatch: ...


class NoOpErrorLogger(ErrorLogger):
    def log_error(
        self,
        error: Exception,
        task: "ReadTask",
        udf: "UDF",
        seq: int,
    ) -> None:
        pass

    def list_errors(self) -> Iterator[str]:
        return iter([])

    def get_error_row(self, error_id: str) -> pa.RecordBatch:
        raise KeyError(f"Error {error_id} not found")


@attrs.define
class CheckpointStoreErrorLogger(ErrorLogger):
    job_id: str = attrs.field()
    checkpoint_store: CheckpointStore = attrs.field()

    def _key(self, task: "ReadTask", seq: int) -> str:
        return f"{self.job_id}/{task.checkpoint_key()}-{seq}"

    def log_error(
        self,
        error: Exception,
        task: "ReadTask",
        udf: "UDF",
        seq: int,
    ) -> None:
        self.checkpoint_store[self._key(task, seq)] = pa.RecordBatch.from_pydict(
            {
                "error": [str(error)],
                "task": [json.dumps(attrs.asdict(task))],
                "seq": [seq],
                "udf": [cloudpickle.dumps(udf)],
            },
        )

    def list_errors(self) -> Iterator[str]:
        yield from self.checkpoint_store.list_keys(f"{self.job_id}/")

    def get_error_row(self, error_id: str) -> pa.RecordBatch:
        return self.checkpoint_store[f"{error_id}"]



import enum
import functools
import hashlib
import inspect
import logging
from collections.abc import Callable
from typing import Any

import attrs
import cloudpickle as pickle
import numpy
import pyarrow as pa

_LOG = logging.getLogger(__name__)


class UDFArgType(enum.Enum):
    """
    The type of arguments that the UDF expects.
    """

    SCALAR = 0
    ARRAY = 1
    RECORD_BATCH = 2


@attrs.define
class UDF(Callable[[pa.RecordBatch], pa.Array]):  # type: ignore
    """User-defined function (UDF) to be applied to a Lance Table."""

    func: Callable = attrs.field()

    name: str = attrs.field()

    cuda: bool = attrs.field(default=False)

    @name.default
    def _name_default(self) -> str:
        if inspect.isfunction(self.func):
            return self.func.__name__
        elif isinstance(self.func, Callable):
            return self.func.__class__.__name__
        else:
            raise ValueError(f"func must be a function or a callable, got {self.func}")

    def _record_batch_input(self) -> bool:
        sig = inspect.signature(self.func)
        if len(sig.parameters) == 1:
            param = list(sig.parameters.values())[0]
            return param.annotation == pa.RecordBatch
        return False

    @property
    def arg_type(self) -> UDFArgType:
        if self._record_batch_input():
            return UDFArgType.RECORD_BATCH
        if _is_batched_func(self.func):
            return UDFArgType.ARRAY
        return UDFArgType.SCALAR

    input_columns: list[str] | None = attrs.field()

    @input_columns.default
    def _input_columns_default(self) -> list[str] | None:
        sig = inspect.signature(self.func)
        params = list(sig.parameters.keys())
        if self._record_batch_input():
            return None
        return params

    @input_columns.validator
    def _input_columns_validator(self, attribute, value) -> None:
        if self.arg_type == UDFArgType.RECORD_BATCH:
            if value is not None:
                raise ValueError(
                    "RecordBatch input UDF must not declare any input column"
                )
            return
        if value is None:
            raise ValueError("Array and Scalar input UDF must declare input column")

    data_type: pa.DataType = attrs.field()

    @data_type.validator
    def _data_type_validator(self, attribute, value) -> None:
        if value is None:
            raise ValueError("data_type must be set")
        if not isinstance(value, pa.DataType):
            raise ValueError(f"data_type must be a pyarrow.DataType, got {value}")

    @data_type.default
    def _data_type_default(self) -> pa.DataType:
        if self.arg_type != UDFArgType.SCALAR:
            raise ValueError(
                "batched UDFs does not support data_type inference yet,"
                " please specify data_type",
            )
        return _infer_func_arrow_type(self.func, None)

    version: str = attrs.field()

    @version.default
    def _version_default(self) -> str:
        # don't use hash(), which is randomly seeded every process startup
        hasher = hashlib.md5()
        # it is fairly safe to to use cloudpickle here because we are using
        # dockerize environments, so the environment should be consistent
        # across all processes
        hasher.update(pickle.dumps(self.func))
        return hasher.hexdigest()

    checkpoint_key: str = attrs.field()

    @checkpoint_key.default
    def _checkpoint_key_default(self) -> str:
        return f"{self.name}:{self.version}"

    def _scalar_func_record_batch_call(self, record_batch: pa.RecordBatch) -> pa.Array:
        """
        WE use this when the UDF uses single call like
        `func(x_int, y_string, ...) -> type`

        this function automatically dispatches rows to the func and returns `pa.Array`
        """

        # this let's us avoid having to allocate a list in python
        # to hold the results. PA will allocate once for us
        def _iter():  # noqa: ANN202
            for item in record_batch.to_pylist():
                # we know input_columns is not none here
                args = [item[col] for col in self.input_columns]  # type: ignore
                yield self.func(*args)

        arr = pa.array(
            _iter(),
            type=self.data_type,
        )
        # this should always by an Array, never should we get a ChunkedArray back here
        assert isinstance(arr, pa.Array)
        return arr

    def __call__(self, *args, **kwargs) -> pa.Array:
        # dispatch coming from Applier or user calling with a `RecordBatch`
        if len(args) == 1 and isinstance(args[0], pa.RecordBatch):
            record_batch = args[0]
            match self.arg_type:
                case UDFArgType.SCALAR:
                    return self._scalar_func_record_batch_call(record_batch)
                case UDFArgType.ARRAY:
                    arrs = [record_batch[name] for name in self.input_columns]  # type: ignore
                    return self.func(*arrs)
                case UDFArgType.RECORD_BATCH:
                    return self.func(record_batch)

        # dispatch is trying to access the function's original pattern
        return self.func(*args, **kwargs)

    field_metadata: dict[str, str] | None = attrs.field(default=None)


def udf(
    func: Callable | None = None,
    *,
    cuda: bool = False,
    data_type: pa.DataType | None = None,
    version: str | None = None,
    field_metadata: dict[str, str] | None = None,
    input_columns: list[str] | None = None,
    **kwargs,
) -> UDF | functools.partial:
    """Decorator of a User Defined Function (UDF).

    Parameters
    ----------
    func: Callable
        The callable to be decorated. If None, returns a partial function.
    data_type: pa.DataType, optional
        The data type of the output PyArrow Array from the UDF.
        If None, it will be inferred from the function signature.
    version: str, optional
        A version string to manage the changes of function.
        If not provided, it will use the hash of the serialized function.
    field_metadata: dict[str, str], optional
        A dictionary of metadata to be attached to the output `pyarrow.Field`.
    input_columns: list[str], optional
        A list of input column names for the UDF. If not provided, it will be
        inferred from the function signature. Or scan all columns.
    arg_type: UDFArgType, optional
        The UDF args type

    """
    if inspect.isclass(func):

        @functools.wraps(func)
        def _wrapper(*args, **kwargs) -> UDF | functools.partial:
            callable_obj = func(*args, **kwargs)
            return udf(
                callable_obj,
                cuda=cuda,
                data_type=data_type,
                version=version,
                field_metadata=field_metadata,
                input_columns=input_columns,
            )

        return _wrapper  # type: ignore

    if func is None:
        return functools.partial(
            udf,
            cuda=cuda,
            data_type=data_type,
            version=version,
            field_metadata=field_metadata,
            input_columns=input_columns,
            **kwargs,
        )

    # we depend on default behavior of attrs to infer the output schema
    def _include_if_not_none(name, value) -> dict[str, Any]:
        if value is not None:
            return {name: value}
        return {}

    args = {
        "func": func,
        "cuda": cuda,
        **_include_if_not_none("data_type", data_type),
        **_include_if_not_none("version", version),
        **_include_if_not_none("field_metadata", field_metadata),
        **_include_if_not_none("input_columns", input_columns),
    }
    # can't use functools.update_wrapper because attrs makes certain assumptions
    # and attributes read-only. We will figure out docs and stuff later
    return UDF(**args)


def _get_annotations(func: Callable) -> dict[str, type]:
    if inspect.isfunction(func):
        return inspect.get_annotations(func)
    elif isinstance(func, Callable):
        return inspect.get_annotations(func.__call__)
    raise ValueError(f"func must be a function or a callable, got {func}")


def _is_batched_func(func: Callable) -> bool:
    annotations = _get_annotations(func)
    if "return" not in annotations:
        return False

    ret_type = annotations["return"]
    if ret_type != pa.Array and not isinstance(ret_type, pa.DataType):
        return False

    input_keys = list(annotations.keys() - {"return"})
    if len(input_keys) == 1:
        return all(
            annotations[input_key] in [pa.RecordBatch, pa.Array]
            for input_key in input_keys
        )

    if any(annotations[input_key] == pa.RecordBatch for input_key in input_keys):
        raise ValueError(
            "UDF can not have multiple parameters with 'pa.RecordBatch' type"
        )
    return all(annotations[input_key] in [pa.Array] for input_key in input_keys)


def _infer_func_arrow_type(func: Callable, input_schema: pa.Schema) -> pa.DataType:
    """Infer the output schema of a UDF

    currently independent of the input schema, in the future we may want to
    infer the output schema based on the input schema, or the UDF itself could
    request the input schema to be passed in.
    """
    if isinstance(func, UDF):
        return func.data_type

    annotations = _get_annotations(func)
    if "return" not in annotations:
        raise ValueError(f"UDF {func} does not have a return type annotation")

    data_type = annotations["return"]
    # do dispatch to handle different types of output types
    # e.g. pydantic -> pyarrow type inference
    if isinstance(data_type, pa.DataType):
        return data_type

    if t := {
        bool: pa.bool_(),
        bytes: pa.binary(),
        float: pa.float32(),
        int: pa.int64(),
        str: pa.string(),
        numpy.bool_: pa.bool_(),
        numpy.bool: pa.bool_(),
        numpy.uint8: pa.uint8(),
        numpy.uint16: pa.uint16(),
        numpy.uint32: pa.uint32(),
        numpy.uint64: pa.uint64(),
        numpy.int8: pa.int8(),
        numpy.int16: pa.int16(),
        numpy.int32: pa.int32(),
        numpy.int64: pa.int64(),
        numpy.float16: pa.float16(),
        numpy.float32: pa.float32(),
        numpy.float64: pa.float64(),
        numpy.str_: pa.string(),
    }.get(data_type):
        return t

    raise ValueError(f"UDF {func} has an invalid return type annotation {data_type}")



from textual.app import App, ComposeResult
from textual.containers import HorizontalScroll, Vertical, VerticalScroll
from textual.widgets import Footer, Header, Static

from geneva.tui._screen import DismissOnQuitScreen


class InfoScreen(DismissOnQuitScreen):
    def __init__(self, *args, msg: str, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.title = "Info"
        self.msg = msg

    def compose(self) -> ComposeResult:
        yield Header(name=self.title)
        yield Footer()
        yield Vertical(
            VerticalScroll(HorizontalScroll(Static(self.msg))),
            Static("q to dismiss", id="dismiss_msg"),
            id="msg_group",
        )

    CSS = """
    InfoScreen {
        background: rgba(128, 128, 128, 0.5);
        align: center middle;
    }
    #dismiss_msg {
        text-align: center;
    }
    #msg_group {
        height: 50%;
        width: 50%;
        background: rgba(64, 74, 89, 1);
        border: dashed white;
    }
    """


def show_info(app: App, *, msg: str) -> None:
    app.push_screen(InfoScreen(msg=msg))



from pathlib import Path

import cloudpickle
import pyarrow as pa
import pyarrow.compute as pc
import pytest

from geneva import InMemoryCheckpointStore, connect, udf
from geneva.apply import LanceRecordBatchUDFApplier, plan_read
from geneva.debug.logger import CheckpointStoreErrorLogger


def test_create_plan(tmp_path: Path) -> None:
    db = connect(tmp_path)
    tbl = db.create_table("tbl", pa.table({"a": [1, 2, 3]}))

    plans = list(plan_read(tbl.uri, batch_size=16))
    assert len(plans) == 1
    plan = plans[0]
    assert plan.uri == tbl.uri
    assert plan.batch_size == 16
    assert plan.offset == 0
    assert plan.limit == 3


@udf(input_columns=["a"])
def one(*args, **kwargs) -> int:
    return 1


def test_applier(tmp_path: Path) -> None:
    db = connect(tmp_path)
    tbl = db.create_table("tbl", pa.table({"a": [1, 2, 3]}))

    plans = list(plan_read(tbl.uri, batch_size=16))
    assert len(plans) == 1
    plan = plans[0]
    assert plan.uri == tbl.uri
    assert plan.batch_size == 16
    assert plan.offset == 0
    assert plan.limit == 3

    store = InMemoryCheckpointStore()
    applier = LanceRecordBatchUDFApplier(
        udfs={"one": one},
        checkpoint_store=store,
    )
    result = applier.run(plan)
    assert len(result) == 1
    batch = pa.RecordBatch.from_pydict(
        {name: store[key]["data"] for name, key in result.items()}
    )
    assert len(batch) == 3
    assert batch.to_pydict() == {"one": [1, 1, 1]}


@udf()
def errors_on_three(a: int) -> int:
    if a == 3:
        raise ValueError("This is an error")
    return 1


def test_applier_error_logging(tmp_path: Path) -> None:
    db = connect(tmp_path)
    tbl = db.create_table("tbl", pa.table({"a": [1, 2, 3]}))

    plans = list(plan_read(tbl.uri, batch_size=16))
    assert len(plans) == 1
    plan = plans[0]
    assert plan.uri == tbl.uri
    assert plan.batch_size == 16
    assert plan.offset == 0
    assert plan.limit == 3

    store = InMemoryCheckpointStore()
    error_logger = CheckpointStoreErrorLogger("job_id", store)
    applier = LanceRecordBatchUDFApplier(
        udfs={
            "one": errors_on_three,
        },
        checkpoint_store=store,
        error_logger=error_logger,
    )
    with pytest.raises(RuntimeError):
        applier.run(plan)

    assert len(list(error_logger.list_errors())) == 1
    error_id = list(error_logger.list_errors())[0]
    error = error_logger.get_error_row(error_id).to_pylist()[0]
    assert error["error"] == "This is an error"
    assert error["udf"] == cloudpickle.dumps(errors_on_three)
    assert error["seq"] == 0


def test_plan_with_filter(tmp_path: Path) -> None:
    db = connect(tmp_path)

    tbl = db.create_table("t", pa.table({"a": range(100)}))
    tbl.add(pa.table({"a": range(100, 200)}))
    tbl.add(pa.table({"a": range(200, 300)}))
    tbl.add(pa.table({"a": range(300, 400)}))

    fragments = tbl.get_fragments()
    assert len(fragments) == 4

    tasks = list(plan_read(tbl.uri, filter="a > 100 AND a % 2 == 0"))
    assert len(tasks) == 3


def test_udf_with_arrow_params(tmp_path: Path) -> None:
    @udf(data_type=pa.int32())
    def batch_udf(a: pa.Array, b: pa.Array) -> pa.Array:
        assert a == pa.array([1, 2, 3])
        assert b == pa.array([4, 5, 6])
        return pc.add(a, b)

    db = connect(tmp_path)
    tbl = db.create_table("t", pa.table({"a": [1, 2, 3], "b": [4, 5, 6]}))

    store = InMemoryCheckpointStore()
    applier = LanceRecordBatchUDFApplier(
        udfs={
            "c": batch_udf,
        },
        checkpoint_store=store,
    )
    result = applier.run(next(plan_read(tbl.uri, batch_size=16)))
    assert len(result) == 1
    batch = pa.RecordBatch.from_pydict(
        {name: store[key]["data"] for name, key in result.items()}
    )
    assert batch == pa.RecordBatch.from_pydict(
        {"c": pa.array([5, 7, 9], type=pa.int32())},
    )



from typing import TYPE_CHECKING

from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.reactive import Reactive, reactive
from textual.widgets import (
    DataTable,
    Footer,
    Header,
    Static,
    TabbedContent,
    TabPane,
    Tree,
)

from geneva.db import Connection
from geneva.table import Table
from geneva.tui._screen import DismissOnQuitScreen, nested
from geneva.tui.error import show_error
from geneva.tui.info import show_info
from geneva.tui.jobs import JobsScreen

if TYPE_CHECKING:
    import pyarrow as pa


class SchemaTable(DataTable):
    BINDINGS = [
        ("ctrl+r", "run", "Run Job"),
    ]

    def __init__(self, *args, table: Table, **kwargs) -> None:
        super().__init__(*args, cursor_type="row", **kwargs)
        self.table = table
        self.add_columns("Name", "Type", "Nullable", "UDF", "Docker")
        for field in self.table.schema:
            udf, docker = "", ""
            field: pa.Field = field
            if (
                field.metadata is not None
                and b"virtual_column.udf_name" in field.metadata
            ):
                udf = field.metadata[b"virtual_column.udf_name"].decode("utf-8")
                docker = field.metadata[b"virtual_column.image"].decode("utf-8")
            self.add_row(field.name, field.type, field.nullable, udf, docker)

    def action_run(self) -> None:
        row = self.cursor_row
        row = self.get_row_at(row)
        name, _, _, udf, docker = row
        if not udf or not docker:
            show_error(self.app, msg="Not a virtual column")
            return
        self.table._conn.jobs.start(table=self.table.name, column=name)
        show_info(self.app, msg="Job started")


class TableDetailsScreen(DismissOnQuitScreen):
    def __init__(
        self,
        *args,
        table: Table,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.title = table.name
        self.table = table

    def compose(self) -> ComposeResult:
        yield Header(name=self.title)
        yield Footer()

        schema = SchemaTable(table=self.table, id="schema_table")

        schema_group = Vertical(
            Static("Schema", id="schema_label"),
            schema,
        )

        indices = list(self.table.list_indices())
        table_info = DataTable(id="table_info_table")
        table_info.add_columns("", "")
        table_info.add_row("Version", self.table.version)
        table_info.add_row("Row Count", self.table.count_rows())
        table_info.add_row("Column Count", len(self.table.schema))
        table_info.add_row("Indices Count", len(indices))

        table_info_group = Vertical(
            Static("Table Info", id="table_info_label"),
            table_info,
        )

        table_detail_group = Horizontal(
            table_info_group,
            schema_group,
        )

        sample = self.table.search().limit(10).to_arrow().to_pylist()

        sample_data = DataTable()
        if sample:
            sample_data.add_columns(*sample[0].keys())
            for row in sample:
                sample_data.add_row(*row.values())

        sample_data_group = Vertical(
            Static("Sample Data", id="sample_data_label"),
            sample_data,
        )

        all_groups = Vertical(
            table_detail_group,
            sample_data_group,
        )

        yield all_groups

    CSS = """
    #schema_table {
        align: center middle;
        width: 100%;
        height: 100%;
    }
    #table_info_table {
        align: center middle;
        width: 100%;
        height: 100%;
    }
    #schema_label {
        text-align: center;
    }
    #table_info_label {
        text-align: center;
    }
    #sample_data_label {
        text-align: center;
    }
    """


class MultiTabTableDetailsScreen(DismissOnQuitScreen):
    BINDINGS = {("r", "refresh", "Refresh")}

    def __init__(
        self,
        *args,
        table: Table,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.title = "Tables"
        self.table = table

    def compose(self) -> ComposeResult:
        yield Header(name=self.title)
        yield Footer()

        with TabbedContent():
            with TabPane("Info", name="Info"):
                # display screen in screen but trim the header and footer
                yield from nested(TableDetailsScreen(table=self.table))
            with TabPane("Job History", name="Job History"):
                yield from nested(
                    JobsScreen(client=self.table._conn.jobs, table_name=self.table.name)
                )

    CSS = """
    #job_history_label {
        text-align: center;
    }
    """
    # Bring in the CSS from the TableDetailsScreen
    CSS += TableDetailsScreen.CSS

    def action_refresh(self) -> None:
        # make sure the connection is still alive
        self.table._conn.table_names()
        self.refresh(recompose=True)


class TablesList(Tree):
    db_conn: Reactive[Connection | None] = reactive(None)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, label="catalog (NOT CONNECTED)", **kwargs)

    def set_connection(self, db_conn: Connection | None) -> None:
        self.db_conn = db_conn
        if self.db_conn is not None:
            self.root.label = "catalog"
            self.root.remove_children()
            self.db = self.root.add_leaf(self.db_conn._uri)
            for table in self.db_conn.table_names():
                self.db.add_leaf(table)
            self.root.expand_all()
        else:
            self.root.label = "catalog (NOT CONNECTED)"
            self.root.remove_children()

    def action_select_cursor(self) -> None:
        super().action_select_cursor()
        node = self.cursor_node
        # don't do anything if the cursor is on the root or the db node
        if node is self.root or node is self.db:
            return

        assert self.db_conn is not None, (
            "if you can select a table, the connection should not be None"
        )

        if node is not None and node.label is not None:
            table = self.db_conn.open_table(str(node.label))
            self.app.push_screen(MultiTabTableDetailsScreen(table=table))



from collections.abc import Callable, Iterator
from typing import Any

from lance.torch.data import LanceDataset, _buffer_arrow_batches, _to_tensor
from torch.utils.data import IterableDataset

from geneva.query import Query
from geneva.table import Table


class Dataset(IterableDataset):
    """PyTorch Dataset for Lance Datalake"""

    def __init__(
        self,
        table: Table,
        batch_size=1,
        to_tensor_fn: Callable | None = None,
    ) -> None:
        self.table = table
        self.batch_size = batch_size
        self._to_tensor_fn = to_tensor_fn or _to_tensor

    # TODO: This annotation sucks
    def __iter__(self) -> Iterator[Any]:
        if isinstance(self.table, Query) or self.table._is_view():
            stream = self.table.to_batches(self.batch_size)
            stream = _buffer_arrow_batches(stream, buffer_size=self.batch_size)
            for batch in stream:
                yield self._to_tensor_fn(batch)
        else:
            for batch in LanceDataset(
                self.table,
                batch_size=self.batch_size,
                to_tensor_fn=self._to_tensor_fn,
            ).__iter__():
                yield batch



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Lance Authors

import functools
import logging
import os
from collections.abc import Callable
from typing import Any
from urllib.parse import urljoin

import attrs
import pyarrow as pa
import requests
from lancedb.common import Credential
from lancedb.remote.errors import LanceDBClientError
from pydantic import BaseModel
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from geneva.remote.timeout import LanceDBClientHTTPAdapterFactory

ARROW_STREAM_CONTENT_TYPE = "application/vnd.apache.arrow.stream"


def _read_ipc(resp: requests.Response) -> pa.Table:
    resp_body = resp.content
    with pa.ipc.open_file(pa.BufferReader(resp_body)) as reader:  # type: ignore
        return reader.read_all()


@attrs.define(slots=False, repr=True)
class RestfulLanceDBClient:
    db_name: str
    region: str
    api_key: Credential
    host_override: str | None = attrs.field(default=None)

    closed: bool = attrs.field(default=False, init=False)

    connection_timeout: float = attrs.field(default=120.0, kw_only=True)
    read_timeout: float = attrs.field(default=300.0, kw_only=True)

    @functools.cached_property
    def session(self) -> requests.Session:
        sess = requests.Session()

        retry_adapter_instance = retry_adapter(retry_adapter_options())
        sess.mount(urljoin(self.url, "/v1/table/"), retry_adapter_instance)

        adapter_class = LanceDBClientHTTPAdapterFactory()
        sess.mount("https://", adapter_class())
        return sess

    @property
    def url(self) -> str:
        return (
            self.host_override
            or f"https://{self.db_name}.{self.region}.api.lancedb.com"
        )

    def __enter__(self) -> "RestfulLanceDBClient":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        self.close()
        return False  # Do not suppress exceptions

    def close(self) -> None:
        self.session.close()
        self.closed = True

    @functools.cached_property
    def headers(self) -> dict[str, str | Credential]:
        headers: dict[str, str | Credential] = {
            "x-api-key": self.api_key,
        }
        if self.region == "local":  # Local test mode
            headers["Host"] = f"{self.db_name}.{self.region}.api.lancedb.com"
        if self.host_override:
            headers["x-lancedb-database"] = self.db_name
        return headers

    @staticmethod
    def _check_status(resp: requests.Response) -> None:
        # Leaving request id empty for now, as we'll be replacing this impl
        # with the Rust one shortly.
        if resp.status_code == 404:
            raise LanceDBClientError(
                f"Not found: {resp.text}", request_id="", status_code=404
            )
        elif 400 <= resp.status_code < 500:
            raise LanceDBClientError(
                f"Bad Request: {resp.status_code}, error: {resp.text}",
                request_id="",
                status_code=resp.status_code,
            )
        elif 500 <= resp.status_code < 600:
            raise LanceDBClientError(
                f"Internal Server Error: {resp.status_code}, error: {resp.text}",
                request_id="",
                status_code=resp.status_code,
            )
        elif resp.status_code != 200:
            raise LanceDBClientError(
                f"Unknown Error: {resp.status_code}, error: {resp.text}",
                request_id="",
                status_code=resp.status_code,
            )

    def get(
        self,
        path: str,
        params: dict[str, Any] | BaseModel | None = None,  # type: ignore
    ) -> str:
        """Send a GET request and returns the deserialized response payload."""
        if isinstance(params, BaseModel):
            params: dict[str, Any] = params.dict(exclude_none=True)
        with self.session.get(
            urljoin(self.url, path),
            params=params,
            headers=self.headers,
            timeout=(self.connection_timeout, self.read_timeout),
        ) as resp:
            self._check_status(resp)
            return resp.json()

    def post(
        self,
        uri: str,
        data: dict[str, Any] | BaseModel | bytes | None = None,  # type: ignore
        params: dict[str, Any] | None = None,
        content_type: str | None = None,
        deserialize: Callable = lambda resp: resp.json(),
        request_id: str | None = None,
    ) -> dict[str, Any]:
        """Send a POST request and returns the deserialized response payload.

        Parameters
        ----------
        uri : str
            The uri to send the POST request to.
        data: Union[dict[str, Any], BaseModel]
        request_id: Optional[str]
            Optional client side request id to be sent in the request headers.

        """
        if isinstance(data, BaseModel):
            data: dict[str, Any] = data.dict(exclude_none=True)
        req_kwargs = {"data": data} if isinstance(data, bytes) else {"json": data}

        headers = self.headers.copy()
        if content_type is not None:
            headers["content-type"] = content_type
        if request_id is not None:
            headers["x-request-id"] = request_id
        with self.session.post(
            urljoin(self.url, uri),
            headers=headers,
            params=params,
            timeout=(self.connection_timeout, self.read_timeout),
            **req_kwargs,  # type: ignore
        ) as resp:
            self._check_status(resp)
            return deserialize(resp)

    def mount_retry_adapter_for_table(self, table_name: str) -> None:
        """
        Adds an http adapter to session that will retry retryable requests to the table.
        """
        retry_options = retry_adapter_options(methods=["GET", "POST"])
        retry_adapter_instance = retry_adapter(retry_options)
        session = self.session

        session.mount(
            urljoin(self.url, f"/v1/table/{table_name}/query/"), retry_adapter_instance
        )
        session.mount(
            urljoin(self.url, f"/v1/table/{table_name}/describe/"),
            retry_adapter_instance,
        )
        session.mount(
            urljoin(self.url, f"/v1/table/{table_name}/index/list/"),
            retry_adapter_instance,
        )


def retry_adapter_options(methods: list[str] | None = None) -> dict[str, Any]:
    if methods is None:
        methods = ["GET"]
    return {
        "retries": int(os.environ.get("LANCE_CLIENT_MAX_RETRIES", "3")),
        "connect_retries": int(os.environ.get("LANCE_CLIENT_CONNECT_RETRIES", "3")),
        "read_retries": int(os.environ.get("LANCE_CLIENT_READ_RETRIES", "3")),
        "backoff_factor": float(
            os.environ.get("LANCE_CLIENT_RETRY_BACKOFF_FACTOR", "0.25")
        ),
        "backoff_jitter": float(
            os.environ.get("LANCE_CLIENT_RETRY_BACKOFF_JITTER", "0.25")
        ),
        "statuses": [
            int(i.strip())
            for i in os.environ.get(
                "LANCE_CLIENT_RETRY_STATUSES", "429, 500, 502, 503"
            ).split(",")
        ],
        "methods": methods,
    }


def retry_adapter(options: dict[str, Any]) -> HTTPAdapter:
    total_retries = options["retries"]
    connect_retries = options["connect_retries"]
    read_retries = options["read_retries"]
    backoff_factor = options["backoff_factor"]
    backoff_jitter = options["backoff_jitter"]
    statuses = options["statuses"]
    methods = frozenset(options["methods"])
    logging.debug(
        f"Setting up retry adapter with {total_retries} retries,"  # noqa G003
        + f"connect retries {connect_retries}, read retries {read_retries},"
        + f"backoff factor {backoff_factor}, statuses {statuses}, "
        + f"methods {methods}"
    )

    return HTTPAdapter(
        max_retries=Retry(
            total=total_retries,
            connect=connect_retries,
            read=read_retries,
            backoff_factor=backoff_factor,
            backoff_jitter=backoff_jitter,
            status_forcelist=statuses,
            allowed_methods=methods,
        )
    )



import os
import traceback

from textual.app import App, ComposeResult
from textual.containers import HorizontalScroll, Vertical, VerticalScroll
from textual.widgets import Footer, Header, Static

from geneva.tui._screen import DismissOnQuitScreen


class ErrorScreen(DismissOnQuitScreen):
    def __init__(self, *args, error_msg: str, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.title = "Error"
        self.msg = error_msg

    def compose(self) -> ComposeResult:
        yield Header(name=self.title)
        yield Footer()
        yield Vertical(
            Static("An error occurred. Please try again later.", id="error"),
            VerticalScroll(HorizontalScroll(Static(self.msg, id="error_msg"))).focus(),
            Static("q to dismiss", id="dismiss_msg"),
            id="error_group",
        )

    CSS = """
    ErrorScreen {
        background: rgba(128, 128, 128, 0.5);
        align: center middle;
    }
    #error_msg {
        color: red;
    }
    #dismiss_msg {
        text-align: center;
    }
    #error {
        color: yellow;
    }
    #error_group {
        height: 50%;
        width: 50%;
        background: rgba(11, 3, 168, 1);
        border: dashed white;
    }
    """


_SHOULD_PDB_ERRORS = os.environ.get("GENEVA_TUI_PDB", "").lower() in (
    "true",
    "1",
    "yes",
    "y",
)


def show_error(
    app: App,
    *,
    msg: str | None = None,
    ex: Exception | None = None,
    start_pdb: bool = _SHOULD_PDB_ERRORS,
) -> None:
    if start_pdb:
        with app.suspend():
            import pdb  # noqa: T100

            pdb.post_mortem()

    if msg is None and ex is None:
        raise ValueError("Either msg or ex must be provided")

    if ex is not None:
        stack = traceback.format_exception(ex)
        msg = "\n".join(stack)

    assert msg is not None

    app.push_screen(ErrorScreen(error_msg=msg))



import logging
import re
import uuid
from collections.abc import Callable, Iterator

import lance
import pandas as pd
import pyarrow as pa

import ray.data
from geneva.apply import LanceRecordBatchUDFApplier, plan_read
from geneva.checkpoint import CheckpointStore
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


def map_with_value(  # noqa: ANN201
    fn: Callable,
):
    def _wrapped(x: dict):  # noqa: ANN202
        result = {
            "task": x["item"],
            "checkpoint_key": fn(x["item"]),
            "frag_id": x["item"].frag_id,
        }
        return result

    return _wrapped


def sort_and_collect_batch(df: pd.DataFrame):  # noqa: ANN201
    data = df.to_dict(orient="records")
    frag_id = data[0]["task"].frag_id
    _LOG.info("Sorting %d keys for frag: %s", len(data), frag_id)
    data.sort(key=lambda x: x["task"].offset)

    checkpoint_keys = [x["checkpoint_key"] for x in data]
    return {"item": [{"frag_id": frag_id, "checkpoint_keys": checkpoint_keys}]}


def _combine_chunks(arr: pa.Array) -> pa.Array:
    if isinstance(arr, pa.ChunkedArray):
        assert len(arr.chunks) == 1
        return arr.chunks[0]
    return arr


def write_fragment(  # noqa: ANN201
    uri: str,
    column_name: str,
    store: CheckpointStore,
):
    dataset = lance.dataset(uri)

    def _write_fragment(x):  # noqa: ANN202
        frag_id = x["item"]["frag_id"]
        checkpoint_keys = x["item"]["checkpoint_keys"]
        batches: list[dict[str, str]] = list(checkpoint_keys)
        import more_itertools

        def _iter() -> Iterator[pa.RecordBatch]:
            for batch in batches:
                res = {
                    key: _combine_chunks(store[value]["data"])
                    for key, value in batch.items()
                }
                yield pa.RecordBatch.from_pydict(res)

        it = more_itertools.peekable(_iter())
        rbr = pa.RecordBatchReader.from_batches(it.peek().schema, it)

        # TODO: this doesn't support struct or complex schema yet
        new_data = lance.fragment.write_fragments(
            rbr,
            uri,
            max_rows_per_file=1 << 31,
            max_bytes_per_file=1 << 40,
        )

        assert len(new_data) == 1
        new_data = new_data[0]
        assert len(new_data.files) == 1
        new_datafile = new_data.files[0]

        # MASSIVE HACK: open up an API to get the field id from the column name
        field_id = re.compile(
            rf'name: "{column_name}", id: (?P<field_id>[\d]*),'
        ).findall(str(dataset.lance_schema))
        assert len(field_id) == 1
        field_id = int(field_id[0])

        new_datafile.fields = [field_id]
        new_datafile.column_indices = [0]

        return {
            "frag_id": frag_id,
            "new_datafile": new_datafile,
        }

    return _write_fragment


def commit_fragments(  # noqa: ANN201
    uri: str,
    version: int,
):
    def _commit_fragments(x) -> None:
        _LOG.info("Committing %d fragments", len(x))
        operation = lance.LanceOperation.DataReplacement(
            replacements=[
                lance.LanceOperation.DataReplacementGroup(
                    fragment_id=item["frag_id"],
                    new_file=item["new_datafile"],
                )
                for item in x
            ]
        )
        lance.LanceDataset.commit(uri, operation, read_version=version)

    return _commit_fragments


def run_ray_add_column(
    uri: str,
    columns: list[str],
    transforms: dict[str, UDF],
    checkpoint_store: CheckpointStore,
    /,
    job_id: str | None = None,
    batch_size: int = 8,
    read_version: int | None = None,
    test_run: bool = True,
    **kwargs,
) -> None:
    if read_version is None:
        read_version = lance.dataset(uri).version

    job_id = job_id or uuid.uuid4().hex

    applier = LanceRecordBatchUDFApplier(
        udfs=transforms,
        checkpoint_store=checkpoint_store,
        error_logger=CheckpointStoreErrorLogger(job_id, checkpoint_store),
    )

    tasks = list(
        plan_read(uri, columns, batch_size=batch_size, read_version=read_version)
    )

    if not len(tasks):
        return None

    output_column = list(transforms.keys())[0]

    ds = (
        ray.data.from_items(tasks)
        .random_shuffle()
        .map(map_with_value(applier.run))
        .groupby("frag_id")
        .map_groups(sort_and_collect_batch)
        .map(write_fragment(uri, output_column, checkpoint_store))
    )
    # trigger collection

    frags = list(ds.iterator().iter_rows())
    _LOG.info("Committing %d fragments", len(frags))
    operation = lance.LanceOperation.DataReplacement(
        replacements=[
            lance.LanceOperation.DataReplacementGroup(
                fragment_id=item["frag_id"],
                new_file=item["new_datafile"],
            )
            for item in frags
        ]
    )
    ds = lance.dataset(uri)
    lance.LanceDataset.commit(uri, operation, read_version=ds.version)



"""Job History Client"""

from collections.abc import Iterable

from geneva.job import Job
from geneva.remote.client import RestfulLanceDBClient


class JobClient:
    """Job History Client"""

    def __init__(self, rest_client: RestfulLanceDBClient) -> None:
        self.rest_client = rest_client

    def list(
        self,
        *,
        table: str | None = None,
        _columns: str | list[str] | None = None,
        _limit: int = 100,
        _offset: int = 0,
    ) -> Iterable[dict]:
        """Get the ID of the Jobs which satisfy the given conditions."""
        req = {"table": table} if table else {}
        resp = self.rest_client.post("/v1/job/list/", data=req)
        yield from resp["jobs"]

    def get_job(self, job_id: str) -> Job: ...

    def upsert_job(self, job: Job) -> None: ...

    def start(
        self,
        table: str,
        column: str,
    ) -> None:
        """Start a new job."""
        self.rest_client.post(
            "/v1/job/start/",
            data={
                "job": {
                    "materialize_virtual_column": {
                        "columns": [column],
                        "table": table,
                    }
                }
            },
            deserialize=lambda x: x,
        )



from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pytest

from geneva import connect, udf
from geneva.transformer import UDF, UDFArgType


@pytest.mark.skip(reason="No support for sync add_columns yet")
def test_udf_fsl(tmp_path: Path) -> None:
    @udf(data_type=pa.list_(pa.float32(), 4))
    def gen_fsl(b: pa.RecordBatch) -> pa.Array:
        arr = pa.array([a * 1.0 for a in range(8)])
        fsl = pa.FixedSizeListArray.from_arrays(arr, 4)
        return fsl

    assert gen_fsl.data_type == pa.list_(pa.float32(), 4)

    db = connect(tmp_path, dataflow_options={"runner": "DirectRunner"})
    tbl = pa.table({"a": [1, 2]})
    tbl = db.create_table("t1", tbl)

    tbl.add_columns({"embed": gen_fsl}, input_columns=["a"])

    tbl = db.open_table("t1")
    assert tbl.schema == pa.schema(
        [
            pa.field("a", pa.int64()),
            pa.field("embed", pa.list_(pa.float32(), 4)),
        ],
    )


def test_udf_data_type_inference() -> None:
    @udf
    def foo(x: int, y: int) -> int:
        return x + y

    assert foo.data_type == pa.int64()
    assert foo.arg_type is UDFArgType.SCALAR

    for np_dtype in [
        np.bool,
        np.bool_,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.float16,
        np.float32,
        np.float64,
    ]:

        @udf
        def foo_np(x: int, np_dtype=np_dtype) -> np_dtype:
            return np_dtype(x)

        assert foo_np.data_type == pa.from_numpy_dtype(np_dtype)
        assert foo_np.arg_type is UDFArgType.SCALAR

    @udf
    def bool_val(x: int) -> bool:
        return x % 2 == 0

    assert bool_val.data_type == pa.bool_()
    assert bool_val.arg_type is UDFArgType.SCALAR

    @udf
    def foo_str(x: int) -> str:
        return str(x)

    assert foo_str.data_type == pa.string()
    assert foo_str.arg_type is UDFArgType.SCALAR

    @udf
    def np_bool(x: int) -> np.bool_:
        return np.bool_(x % 2 == 0)

    assert np_bool.data_type == pa.bool_()
    assert np_bool.arg_type is UDFArgType.SCALAR


def test_udf_as_regular_functions() -> None:
    @udf
    def add_three_numbers(a: int, b: int, c: int) -> int:
        return a + b + c

    assert add_three_numbers(1, 2, 3) == 6
    assert add_three_numbers(10, 20, 30) == 60
    assert add_three_numbers.arg_type is UDFArgType.SCALAR
    assert add_three_numbers.data_type == pa.int64()

    @udf
    def make_string(x: int, y: str) -> str:
        return f"{y}-{x}"

    assert make_string(42, "answer") == "answer-42"
    assert make_string.arg_type is UDFArgType.SCALAR
    assert make_string.data_type == pa.string()

    @udf(data_type=pa.float32())
    def multi_by_two(batch: pa.RecordBatch) -> pa.Array:
        arr = pc.multiply(batch.column(0), 2)
        return arr

    rb = pa.RecordBatch.from_arrays([pa.array([1, 2, 3])], ["col"])
    assert multi_by_two(rb) == pa.array([2, 4, 6])
    assert multi_by_two.arg_type is UDFArgType.RECORD_BATCH

    # Confirm direct calls with multiple arguments still work as expected
    assert make_string(7, "num") == "num-7"
    assert add_three_numbers(2, 3, 4) == 9


def test_udf_with_batch_mode() -> None:
    """Test using a scalar UDF, but filled with batch model"""

    @udf
    def powers(a: int, b: int) -> int:
        return a**b

    # a RecordBatch with a and b columns
    rb = pa.RecordBatch.from_arrays(
        [pa.array([1, 2, 3]), pa.array([4, 5, 6])],
        ["a", "b"],
    )
    result = powers(rb)
    assert result == pa.array([1, 2**5, 3**6])


def test_stateful_callable() -> None:
    @udf
    class StatefulFn:
        def __init__(self) -> None:
            self.state = 0

        def __call__(self, x: int) -> int:
            self.state += x
            return self.state

    stateful_fn = StatefulFn()
    assert isinstance(stateful_fn, UDF)
    assert stateful_fn(1) == 1
    assert stateful_fn.arg_type is UDFArgType.SCALAR
    assert stateful_fn.data_type == pa.int64()
    assert stateful_fn.input_columns == ["x"]

    @udf(data_type=pa.int64())
    class SatefulBatchFn:
        def __init__(self) -> None:
            self.state = 0

        def __call__(self, batch: pa.RecordBatch) -> pa.Array:
            self.state += sum(batch.column(0).to_pylist())
            return pa.array([self.state] * batch.num_rows)

    stateful_batch_fn = SatefulBatchFn()
    assert isinstance(stateful_batch_fn, UDF)
    assert stateful_batch_fn.arg_type is UDFArgType.RECORD_BATCH
    assert stateful_batch_fn.data_type == pa.int64()


def test_batched_udf_with_explicity_columns() -> None:
    @udf(data_type=pa.int64())
    def add_columns(a: pa.Array, b: pa.Array) -> pa.Array:
        return pc.add(a, b)

    assert add_columns.arg_type is UDFArgType.ARRAY
    assert add_columns.data_type == pa.int64()
    assert add_columns.input_columns == ["a", "b"]

    with pytest.raises(
        ValueError, match="multiple parameters with 'pa.RecordBatch' type"
    ):

        @udf
        def bad_udf(a: pa.RecordBatch, b: pa.RecordBatch) -> pa.Array:
            return pc.add(a.column(0), b.column(0))
